# -*- coding: utf-8 -*-

import os
import time
from threading import Lock

import requests
from flask import Flask as _Flask

from . import patch
from . import constants
from .mocks import BaseMock
from .tools import endpoint
from .client import MockServerClient
from .exceptions import MockServerError
from .mocks import get_mock_class_by_file_name
from .tools import set_log_level_for_dependency
from .mocks import on_file as fill_mock_from_file


MOCK_FILE_EXTENSION = '.mock'


# Patch flask
Flask = type(
    'Flask',
    (_Flask, ),
    {
        'add_url_rule': patch.add_url_rule,
        'dispatch_request': patch.dispatch_request(
            _Flask.dispatch_request,
        ),
    },
)


lock = Lock()


class MockServer(object):

    __client_class__ = MockServerClient

    def __init__(self, config):
        """
        :type config: seismograph.ext.mocker.extension.Config
        """
        set_log_level_for_dependency(config.DEBUG)

        self.__app = Flask(
            constants.WSGI_APP_NAME,
            static_url_path=config.STATIC_URL_PATH,
            static_folder=config.STATIC_FOLDER
        )
        self.__app.debug = config.DEBUG

        self.__config = config
        self.__debug = config.DEBUG

        self._process = None
        self._started = False
        self._client = self.__client_class__(config)

        if config.PATH_TO_MOCKS:
            self.scan_dir(config.PATH_TO_MOCKS)

        with self.__app.app_context():
            from .api import MockerApi
            MockerApi(self).install()

    def __call__(self, *args, **kwargs):
        return self.wsgi_handler(*args, **kwargs)

    @property
    def _stopped(self):
        return not self._started

    @property
    def views(self):
        return self.__app.view_functions

    @property
    def config(self):
        return self.__config

    @property
    def client(self):
        return self._client

    @property
    def debug(self):
        return self.__debug

    @debug.setter
    def debug(self, value):
        self.__debug = value
        self.__app.debug = value
        set_log_level_for_dependency(value)

    @property
    def mocks(self):
        return (
            v for v in self.views.values()
            if isinstance(v, BaseMock)
        )

    @property
    def wsgi_handler(self):
        return self.__app.wsgi_app

    def scan_dir(self, mock_path):
        for name in os.listdir(mock_path):
            path = os.path.join(mock_path, name)

            if name.endswith(MOCK_FILE_EXTENSION):
                with open(path, 'r') as fp:
                    mock_cls = get_mock_class_by_file_name(name)
                    mock = mock_cls()
                    fill_mock_from_file(mock, fp)
                self.add_mock(mock)

            if os.path.isdir(path):
                self.scan_dir(path)

    def add_url_rule(self, rule, endpoint=None, view_func=None, **options):
        with lock:
            self.__app.add_url_rule(
                rule, endpoint=endpoint, view_func=view_func, **options
            )

    def add_mock(self, mock):
        """
        :type mock: BaseMock
        :type rewrite: bool
        """
        self.add_url_rule(
            mock.url_rule,
            view_func=mock,
            endpoint=endpoint(
                mock.url_rule,
                mock.http_method,
            ),
            methods=[mock.http_method],
        )

    def run(self):
        self.__app.run(host=self.__config.HOST, port=self.__config.PORT)

    def serve_forever(self):
        if self.__config.GEVENT:
            from gevent.wsgi import WSGIServer

            WSGIServer(
                (self.__config.HOST, self.__config.PORT), self,
                log='default' if self.__debug else None,
            ).serve_forever()
        else:
            from werkzeug.serving import ThreadedWSGIServer

            ThreadedWSGIServer(self.__config.HOST, self.__config.PORT, self).serve_forever()

    def waiting_for_ready(self, timeout=constants.DEFAULT_START_SERVER_TIMEOUT):
        if self._stopped:
            raise MockServerError('Mock server is not started')

        is_error = True
        t_start = time.time()

        while time.time() <= t_start + timeout:
            if not is_error:
                break

            try:
                self._client.get('/')
                is_error = False
            except requests.ConnectionError:
                pass

        if is_error:
            raise MockServerError(
                'Mock server has not been started for "{}" sec.'.format(timeout),
            )

    def start(self):
        if self._started:
            return

        if self.__config.GEVENT:
            from .runners.gevent import RunServerInGreenlet

            self._process = RunServerInGreenlet(self)
        else:
            from werkzeug.serving import ThreadedWSGIServer
            from .runners.threading import RunServerInThread

            self._process = RunServerInThread(self, ThreadedWSGIServer)

        self._process.start()

        self._started = True
        self.waiting_for_ready()

    def restart(self):
        self.stop()
        self.start()

    def stop(self):
        if self._stopped:
            return

        if self.__config.GEVENT:
            self._process.kill()
        else:
            self._process.stop()

        self._process.join()

        self._process = None
        self._started = False



# -*- coding: utf-8 -*-

import sys

try:
    from StringIO import StringIO
except ImportError:  # please python 3
    from io import StringIO

from contextlib import contextmanager


@contextmanager
def dev_null():
    class MockStd(object):

        def __getattr__(self, item):
            def null(*args, **kwargs):
                pass
            return null

    stdout = sys.stdout
    stderr = sys.stderr
    sys.stdout = MockStd()
    sys.stderr = MockStd()
    try:
        yield
    finally:
        sys.stdout = stdout
        sys.stderr = stderr


@contextmanager
def capture_stdout(stream=None):
    stream = stream or StringIO()
    stdout = sys.stdout
    sys.stdout = stream
    try:
        yield stream
    finally:
        sys.stdout = stdout


@contextmanager
def capture_stderr(stream=None):
    stream = stream or StringIO()
    stderr = sys.stderr
    sys.stderr = stream
    try:
        yield stream
    finally:
        sys.stderr = stderr


@contextmanager
def capture_output(stream=None):
    stream = stream or StringIO()
    with capture_stdout(stream):
        with capture_stderr(stream):
                yield stream



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import calendar
import datetime as _dt
from functools import wraps
from dateutil.relativedelta import relativedelta


DEFAULT_DATE_FORMAT = '%Y-%m-%d'
DEFAULT_DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S'


def _make_copy(instance, cls=None):
    if isinstance(instance, _dt.datetime):
        cls = cls or _dt.datetime
        return cls(
            instance.year,
            instance.month,
            instance.day,
            instance.hour,
            instance.minute,
            second=instance.second,
            microsecond=instance.microsecond,
            tzinfo=instance.tzinfo,
        )
    elif isinstance(instance, _dt.date):
        cls = cls or _dt.date
        return cls(
            instance.year,
            instance.month,
            instance.day,
        )

    raise TypeError(
        'Incorrect type of object {} Expected "datetime.date" or "datetime.datetime"'.format(instance),
    )


def _make_result(data, *handlers):
    for handler in handlers:
        data = handler(data)

    return data


def to_string(d, fmt=None):
    default_fmt = (
        DEFAULT_DATETIME_FORMAT
        if isinstance(d, _dt.datetime)
        else
        DEFAULT_DATE_FORMAT
    )
    return d.strftime(fmt or default_fmt)


def date_args_to_string(fmt):
    def wrapper(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            new_args = []
            for a in args:
                if isinstance(a, _dt.date):
                    new_args.append(to_string(a, fmt))
                else:
                    new_args.append(a)

            for k, v in kwargs.items():
                if isinstance(v, _dt.date):
                    kwargs[k] = to_string(v, fmt)
            return f(*new_args, **kwargs)

        return wrapped
    return wrapper


def minus_delta(datetime, **kwargs):
    return _make_copy(datetime - _dt.timedelta(**kwargs))


def plus_delta(datetime, **kwargs):
    return _make_copy(datetime + _dt.timedelta(**kwargs))


def minus_seconds(datetime, seconds, *handlers):
    return _make_result(minus_delta(datetime, seconds=seconds), *handlers)


def plus_seconds(datetime, seconds, *handlers):
    return _make_result(plus_delta(datetime, seconds=seconds), *handlers)


def minus_minutes(datetime, minutes, *handlers):
    return _make_result(minus_delta(datetime, minutes=minutes), *handlers)


def plus_minutes(datetime, minutes, *handlers):
    return _make_result(plus_delta(datetime, minutes=minutes), *handlers)


def minus_hours(datetime, hours, *handlers):
    return _make_result(minus_delta(datetime, hours=hours), *handlers)


def plus_hours(datetime, hours, *handlers):
    return _make_result(plus_delta(datetime, hours=hours), *handlers)


def minus_days(datetime, days, *handlers):
    return _make_result(minus_delta(datetime, days=days), *handlers)


def plus_days(datetime, days, *handlers):
    return _make_result(plus_delta(datetime, days=days), *handlers)


def minus_weeks(datetime, weeks, *handlers):
    return _make_result(minus_delta(datetime, weeks=weeks), *handlers)


def plus_weeks(datetime, weeks, *handlers):
    return _make_result(plus_delta(datetime, weeks=weeks), *handlers)


def minus_months(datetime, months, *handlers):
    return _make_result(_make_copy(datetime - relativedelta(months=months)), *handlers)


def plus_months(datetime, months, *handlers):
    return _make_result(_make_copy(datetime - relativedelta(months=-months)), *handlers)


def minus_years(datetime, years, *handlers):
    return _make_result(_make_copy(datetime - relativedelta(years=years)), *handlers)


def plus_years(datetime, years, *handlers):
    return _make_result(_make_copy(datetime - relativedelta(years=-years)), *handlers)


def to_start_month(datetime, *handlers):
    return _make_result(_make_copy(datetime.replace(day=1)), *handlers)


def to_start_year(datetime, *handlers):
    return _make_result(_make_copy(datetime.replace(day=1, month=1)), *handlers)


def to_end_month(datetime, *handlers):
    _, end_day = calendar.monthrange(datetime.year, datetime.month)
    return _make_result(_make_copy(datetime.replace(day=end_day)), *handlers)


def to_date(date, *handlers):
    return _make_result(_dt.date(date.year, date.month, date.day), *handlers)


def now(*handlers):
   return _make_result(_dt.datetime.now(), *handlers)


def date(year, month, day, *handlers):
    return _make_result(_dt.date(year, month, day), *handlers)


def today(*handlers):
    return _make_result(_dt.date.today(), *handlers)


class delta:

    minus_seconds = staticmethod(lambda s: lambda d: minus_seconds(d, s))
    plus_seconds = staticmethod(lambda s: lambda d: plus_seconds(d, s))
    minus_minutes = staticmethod(lambda m: lambda d: minus_minutes(d, m))
    plus_minutes = staticmethod(lambda m: lambda d: plus_minutes(d, m))
    minus_hours = staticmethod(lambda h: lambda d: minus_hours(d, h))
    plus_hours = staticmethod(lambda h: lambda d: plus_hours(d, h))
    minus_days = staticmethod(lambda y: lambda d: minus_days(d, y))
    plus_days = staticmethod(lambda y: lambda d: plus_days(d, y))
    minus_weeks = staticmethod(lambda w: lambda d: minus_weeks(d, w))
    plus_weeks = staticmethod(lambda w: lambda d: plus_weeks(d, w))
    minus_months = staticmethod(lambda m: lambda d: minus_months(d, m))
    plus_months = staticmethod(lambda m: lambda d: plus_months(d, m))
    minus_years = staticmethod(lambda y: lambda d: minus_years(d, y))
    plus_years = staticmethod(lambda y: lambda d: plus_years(d, y))


class fmt:

    to_string = staticmethod(
        lambda f=None: lambda d: to_string(d, f or DEFAULT_DATE_FORMAT)
    )
